#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
* Filename: __init__.py.py
* Description:
* Time: 2020.11.25
* Author: liuf5
*/
"""

__name__ = 'zoomeye'
__package__ = 'zoomeye'
__version__ = ' 2.0.4.6.11'
__site__ = "https://www.zoomeye.org/doc"
